/*  QUALCOMM Proprietary                                                      * 
 *                                                                            * 
 *  Copyright (C) 2012 QUALCOMM Incorporated.  All rights reserved.           * 
 *                                                                            * 
 *                                                                            * 
 *  QUALCOMM Incorporated provides this material "AS IS" and without          *  
 *  warranty of any kind. Any implied or statutory warranties, including, but * 
 *  not limited to, the implied warranties of merchantability,                * 
 *  noninfringement or fitness for a particular purpose, are disclaimed.      * 
 *                                                                            * 
 *  This technology was exported from the United States in accordance         * 
 *  with the Export Administration Regulations. Diversion contrary to         * 
 *  U.S. law prohibited.
*/
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifndef _WIN32
#include <unistd.h>
#include <netinet/in.h>
#include <inttypes.h>
#else
#include <Winsock2.h>
#endif


#define MIN(a,b)  (((a)<(b))?(a):(b))
#define MAX(x,y)  (((x)>(y))?(x):(y))
//#define DIAGNOSTICS

#define PREALLOCATE_SIZE 25000
#define MAX_NUM_BITS  10000
#define G192_SYNC_WORD   27425
#define ACTIVE_FRAME     1
#define SID_FRAME        2
#define NODATA_FRAME     0


typedef struct{
    int length;                      /* length of the RTP payload and RTP header                                 */
    long frame_rate;                 /* frame rate parameter: can be of ACTIVE_FRAME, SID_FRAME or NODATA_FRAME  */
    short payload[MAX_NUM_BITS];  /* the payload                                                              */
} PAYLOAD_INFO ;

typedef struct{
    int rtp_sn;
    long rtp_ts;
    long pkt_ind;
    double arrival_time;
} PROF_STAT;

/*-------------------------------------------------------------------------- *
*  Read the G.192 EVS Packet and derive frame size and type from it          *
*--------------------------------------------------------------------------  */

int read_packet(FILE *fp_evspkt, short *framesize, short *frametype, short *payload)
{

    short val=0;
    int count;
    
    /* Read the sync word */
    count=fread(&val,sizeof(short),1,fp_evspkt);

    if(count == 0)                                        /* Done reading all packets */
    {
        return(0);
    }

    if(val != G192_SYNC_WORD)
    {
        fprintf(stderr, "G192 SYNC WORD CORRUPTED\n");
        return(0);
    }

    /* Read the framesize */
    fread(&val,sizeof(short),1,fp_evspkt);

    switch(val)
    {
    case 0:
        *frametype = NODATA_FRAME;
        break;
    case 35:                                             /* This represents the AMR-WB SID size, candidate should modify */
        *frametype = SID_FRAME;                          /* to match its SID frame size                                  */  
        break;
    default:
        *frametype = ACTIVE_FRAME;
    }

    /* The framesize should include the syncword and the framesize */
    *framesize = val + 2;

    /* Read the vocoder frame and store in payload in the G.192 format */
    payload[0] = G192_SYNC_WORD;
    payload[1] = val;
    fread(payload + 2, sizeof(short), val, fp_evspkt);
    
    return(1);
	
}

/*-------------------------------------------------------------------------------*
*  Comparator function for quick sort: sorts in increasing order of receive time
*--------------------------------------------------------------------------------*/

int arrivalcompare (const void *a, const void *b)
{
    int retval;
    PROF_STAT *tmp1, *tmp2;
    tmp1=(PROF_STAT *)a;
    tmp2=(PROF_STAT *)b;
    retval=((int)(tmp1->arrival_time - tmp2->arrival_time));

    /* If the two packets have the same arrival time due to bundling, then use the temporary
       rtp sequence number to order the packets correctly */

    if(retval == 0)                    
    {
        retval=tmp1->pkt_ind - tmp2->pkt_ind;
    }
    return(retval);
}

/*--------------------------------------------------------------------------*
*  main function call for network simulator                                *
*--------------------------------------------------------------------------*/

int main (int argc, char **argv)
{

    FILE *fp_delayprof;                /* Delay loss profile file     */
    FILE *fp_encpkt;                   /* Encoder packet file         */
    FILE *fp_output;                   /* RTP packet output file      */

    PAYLOAD_INFO *payload_array=NULL;
    int payload_array_size = 0;        /*size of payload array, may contain empty elements */
    int num_pkts = 0;                  /* numb of payload, i.e., num of RTP packet */
    int idx, pkt_cnt;
    long framerate;
    int dtx_val=0;
    int fsamp;

    short frame_type;
    short frame_size;
    short tmp_payload[MAX_NUM_BITS];

    int delaytemp;
    long prev_sn, prev_ts;

    int rtp_sn_BE;
    long rtp_ts_BE, timestmpseed;

    int smallest_rtp_sn = 10000000;
    int last_rtp_sn = -1;

    long size;
    int i, j, k, arrival_time_int, vpcount; 
    
    short rtp_header_field1 = 22;
    long rtp_header_field2 = 22;
    int num_frm_per_pkt = 1;
    PROF_STAT *dlprof;
    int dlprof_size, tmp, firstAftND;



    fprintf(stdout, "\n");
    fprintf(stdout, "*******************************************************************************\n");
    fprintf(stdout, "          Network simulator for creating input to the voip decoder             \n");
    fprintf(stdout, "*******************************************************************************\n");
    fprintf(stdout, "\n");


    /* open files for reading and writing */
    if (argc < 5)
    {
	fprintf(stderr,"Usage: %s Fs Delay_Loss_Profile bitstream_file Output_file Output_seq_file <optional: num_frames_per_pkt> \n",argv[0]);
        fprintf(stdout,"Required arguments:\n");
        fprintf(stdout,"Fs                 : Sampling frequency in Hz (8000, 16000, 32000 or 48000)\n");
        fprintf(stdout,"Delay_Loss_Profile : Delay loss profile\n");
        fprintf(stdout,"bitstream_file     : Input bitstream filename \n");
        fprintf(stdout,"output_file        : Output file to be used as input to ncim \n\n");
        fprintf(stdout,"Options:\n");
        fprintf(stdout, "num_frames_per_pkt: Number of frames per packet; default is 1\n");
        exit(0);
    }
    fsamp = (int) atoi( argv[1] );

    if ((fp_delayprof = fopen (argv[2], "r")) == NULL )
    {
        fprintf(stderr, "ERROR: Can not open %s \n", argv[2]);
        exit(-1);
    }

    if ((fp_encpkt = fopen (argv[3], "rb") ) == NULL )
    {
        fprintf(stderr, "ERROR: Can not open %s \n", argv[3]);
        exit(-1);
    }

    if ((fp_output = fopen (argv[4], "wb")) == NULL )
    {
        fprintf (stderr, "ERROR: Output File Not Opened \n");
        exit(-1);
    }

    if(argc == 6)
    {
        num_frm_per_pkt = (int) atoi (argv[5]);
    }
    num_pkts = 0;

    /* Read and store all packets in memory */
    while (read_packet(fp_encpkt,&frame_size, &frame_type, tmp_payload))
     {
        /* Store the packet in memory */
        if (num_pkts+1 > payload_array_size) 
        {
            payload_array = (PAYLOAD_INFO*) realloc(payload_array, sizeof(PAYLOAD_INFO) *  (PREALLOCATE_SIZE+payload_array_size));
            if (!payload_array) 
            {
                fprintf(stderr, "ERROR to reallocate memory \n");
                exit(-1);
            }
            payload_array_size += PREALLOCATE_SIZE;
        }

        payload_array[num_pkts].frame_rate = frame_type;
        payload_array[num_pkts].length = frame_size;

        memcpy(payload_array[num_pkts].payload, tmp_payload, frame_size*sizeof(short));
        num_pkts++;

    }

	
    
    /* Randomize the starting time as required in TS 26.114 */
    srand((unsigned)time(NULL));
    timestmpseed = (long)(rand()%10000);
    timestmpseed = 0;

    /* Read the delay loss profile */
	
	dlprof_size = 0;
	while(!feof(fp_delayprof))
	{
		fscanf(fp_delayprof,"%d",&tmp);
		dlprof_size++;
	}
	dlprof_size--;
	dlprof_size = MIN(dlprof_size * num_frm_per_pkt , num_pkts); 
	dlprof = (PROF_STAT *) malloc(( dlprof_size + num_frm_per_pkt )*sizeof(PROF_STAT));

	fseek (fp_delayprof, 0, SEEK_SET);

    /* Simulate the RTP time stamp, sequence number and the arrival time based on the read profile */
    i=0; vpcount = 0; 
    while (!feof(fp_delayprof) && i < dlprof_size )
    {
        fscanf(fp_delayprof, "%d",&delaytemp);
        for (j=1; j <= num_frm_per_pkt; j++)
        {
            dlprof[vpcount].pkt_ind = i+j;
            dlprof[vpcount].rtp_ts = timestmpseed + (long)(((float) fsamp/8000)* 160);
            if (j == num_frm_per_pkt && delaytemp != -1)
            {
                dlprof[vpcount].arrival_time= ((8000.0/(float)fsamp)*(float)(dlprof[vpcount].rtp_ts)/8.0)+delaytemp;
                for (k=1;k<num_frm_per_pkt;k++)
                {
                    dlprof[vpcount-k].arrival_time = dlprof[vpcount].arrival_time;
                    dlprof[vpcount-k].rtp_ts = dlprof[vpcount].rtp_ts; 
                }
            }

            timestmpseed = dlprof[vpcount].rtp_ts;
            if (delaytemp != -1)
            {
                vpcount++;
            }
        }
        i= i + num_frm_per_pkt;
    }
    vpcount--;
    
    smallest_rtp_sn = 1;


    /* Sort the information according to increasing arrival time*/
    qsort(dlprof,vpcount,sizeof(PROF_STAT),arrivalcompare);


    /* Write the packets to the output */
    pkt_cnt= 0;
    prev_ts=0;
    prev_sn=0;
    firstAftND=0;

    for(i=0;i<vpcount;i++)
    {
         
        last_rtp_sn = dlprof[i].pkt_ind;
        arrival_time_int = (int) dlprof[i].arrival_time;
        idx = dlprof[i].pkt_ind - smallest_rtp_sn;
        if (idx - num_pkts > 10 ) break;
        else if (idx >= num_pkts ) continue;

        size = payload_array[idx].length + 12;                                          /* rtp header + the size of evs payload (in bytes) */
	    framerate = payload_array[idx].frame_rate;

        tmp= (long) ((float)dlprof[i].rtp_ts*((float) 8000/(float) fsamp)/(float)160);
        dlprof[i].rtp_sn = prev_sn + (int)((float) (tmp - prev_ts)/(float)(num_frm_per_pkt));
        prev_ts = tmp;
     
        
        if (framerate == NODATA_FRAME)                              /* frames with no data */
        {                                 
            dtx_val = 1;
            firstAftND = 1;
        }
        if (dtx_val) 
        {
            dtx_val = 0;
            continue;
        }
        if(firstAftND == 1 &&( dlprof[i].rtp_sn == prev_sn))
        {
            dlprof[i].rtp_sn = dlprof[i].rtp_sn + 1;
        }
        firstAftND=0;
        prev_sn = dlprof[i].rtp_sn;

        
	    fwrite (&size, sizeof (long), 1, fp_output);                  /*  Write the Packet Size including the RTP Header size */
	    fwrite (&arrival_time_int, sizeof (long), 1, fp_output);      /*  Write the Packet Arrival Time                       */

/*      The following RTP header fields are now added (refer to RFC 3550
*       for details:
*  
*       0                   1                   2                   3
*       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
*      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*      |V=2|P|X|  CC   |M|     PT      |       sequence number         |
*      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*      |                           timestamp                           |
*      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*      |           synchronization source (SSRC) identifier            |
*      +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
*      |            contributing source (CSRC) identifiers             |
*      |                             ....                              |
*      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ 
*/ 

	    /* RTP Header - 12 bytes */
        fwrite (&rtp_header_field1, sizeof (short), 1, fp_output);    /*  Write 2 bit V, 1 bit P, 1 bit X, 4 bit CC,
                                                                                1 bit M, 7 bit PT = Total of 16 bits           */
                                                                        
        rtp_sn_BE = (htons (dlprof[i].rtp_sn));                       
        fwrite (&rtp_sn_BE, sizeof (short), 1, fp_output);            /*  Write the 16 bit RTP sequence number                 */

        rtp_ts_BE = htonl (dlprof[i].rtp_ts);                         
        fwrite (&rtp_ts_BE, sizeof (long), 1, fp_output);             /*  Write the 32 bit RTP time stamp                      */

        fwrite (&rtp_header_field2, sizeof (long), 1, fp_output);     /*  Write the 32 bit SSRC sequence number                */

	    /* RTP Payload */

        /* Write the RTP payload (in the G.192 format) */
		fwrite (payload_array[idx].payload, sizeof (short), payload_array[idx].length, fp_output);

#ifdef DIAGNOSTICS
        fprintf(stdout,"%d %d %f %d\n", dlprof[i].rtp_sn, dlprof[i].rtp_ts, dlprof[i].arrival_time, payload_array[idx].frame_rate);
#endif
        pkt_cnt ++;


    }
    fclose(fp_output);
    fclose(fp_delayprof);
    fclose(fp_encpkt);

    free(dlprof);
	free(payload_array);
    return(1);

}
